#include <iostream>
#include <stdlib.h>
#include <string.h>

#include "pyanim.h"

video::video(const char *filename, unsigned w, unsigned h, 
             int framerate, int bitrate) 
{
  width = w;
  height = h;
  
  av_register_all();
  
  fmt = guess_format(NULL, filename, NULL);
  if (!fmt) {
    std::cerr << "Could not deduce output format from"
              << " file extension: using MPEG: \"" << filename << "\""
              << std::endl;
    fmt = guess_format("mpeg", NULL, NULL);
  }
  
  oc = av_alloc_format_context();
  if (!oc) {
    std::cerr << "Memory error\n" << std::endl;
    exit(1);
  }
  
  oc->oformat = fmt;
  snprintf(oc->filename, sizeof(oc->filename), "%s", filename);
  
  video_st = NULL;
  
  if (fmt->video_codec != CODEC_ID_NONE) {
    video_st = add_video_stream(oc, fmt->video_codec, framerate, bitrate);
  }
  
  if (av_set_parameters(oc, NULL) < 0) {
    std::cerr << "Invalid output format parameters" << std::endl;
    exit(1);
  }
  
  // start file
  create(filename);
}

video::~video() {
  /* free the streams */
  for(int i = 0; i < oc->nb_streams; i++) {
    av_freep(&oc->streams[i]);
  }
  
  if (!(fmt->flags & AVFMT_NOFILE)) {
    /* close the output file */
    url_fclose(&oc->pb);
  }
  
  /* free the stream */
  av_free(oc);
}

AVStream *video::add_video_stream(AVFormatContext *octx, int codec_id,
                                  int framerate, int bitrate) 
{
  AVCodecContext *ctx;
  AVStream *stream;
  
  stream = av_new_stream(octx, 0);
  if (!stream) {
    std::cerr << "Could not allocate stream" << std::endl;
  }
  
  ctx = &stream->codec;
  ctx->codec_id = (CodecID)codec_id;
  ctx->codec_type = CODEC_TYPE_VIDEO;
  ctx->bit_rate = bitrate;
  ctx->width = width;
  ctx->height = height;
  ctx->frame_rate = framerate;
  
  // not sure what these do...
  ctx->frame_rate_base = 1;
  ctx->gop_size = 12;
  
  if (!strcmp(octx->oformat->name, "mp4") ||
      !strcmp(octx->oformat->name, "mov") ||
      !strcmp(octx->oformat->name, "3gp"))
    ctx->flags |= CODEC_FLAG_GLOBAL_HEADER;
  
  return stream;
}

void video::create(const char *filename) {
  dump_format(oc, 0, filename, 1);
  open_video();    
  
  if (!(fmt->flags & AVFMT_NOFILE)) {
    if (url_fopen(&oc->pb, filename, URL_WRONLY) < 0) {
      fprintf(stderr, "Could not open '%s'\n", filename);
      exit(1);
    }
  }
  
  /* write the stream header, if any */
  av_write_header(oc);
}


AVFrame *video::alloc_picture(int pix_fmt, int width, int height) {
  AVFrame *picture;
  uint8_t *picture_buf;
  int size;
    
  picture = avcodec_alloc_frame();
  if (!picture) return NULL;

  size = avpicture_get_size(pix_fmt, width, height);
  picture_buf = new uint8_t[size];
  if (!picture_buf) {
    av_free(picture);
    return NULL;
  }
  avpicture_fill((AVPicture *)picture, picture_buf, 
                 pix_fmt, width, height);
  return picture;
}

void video::open_video() {
  AVCodec *codec;
  AVCodecContext *c;

  c = &video_st->codec;
  
  /* find the video encoder */
  codec = avcodec_find_encoder(c->codec_id);
  if (!codec) {
    std::cerr << "codec not found" << std::endl;
    exit(1);
  }
  
  /* open the codec */
  if (avcodec_open(c, codec) < 0) {
    std::cerr << "could not open codec" << std::endl;
    exit(1);
  }
  
  video_outbuf = NULL;
  if (!(oc->oformat->flags & AVFMT_RAWPICTURE)) {
    /* allocate output buffer */

    // FIXME: How to determine this? No docs found...
    video_outbuf_size = 200000;
    //video_outbuf_size = width*height*video_st->codec.pix_fmt;
    video_outbuf = new uint8_t[video_outbuf_size];
  }

  /* allocate the encoded raw picture */
  picture = alloc_picture(c->pix_fmt, c->width, c->height);
  if (!picture) {
    std::cerr << "Could not allocate picture" << std::endl;
    exit(1);
  }

  /* if the output format is not YUV420P, then a temporary YUV420P
     picture is needed too. It is then converted to the required
     output format */
  tmp_picture = NULL;
  if (c->pix_fmt != PIX_FMT_RGB24) {
    tmp_picture = alloc_picture(PIX_FMT_RGB24, c->width, c->height);
    if (!tmp_picture) {
      std::cerr << "Could not allocate temporary picture" << std::endl;
      exit(1);
    }
  }
}

void video::close_video(AVFormatContext *oc, AVStream *st) {
  avcodec_close(&video_st->codec);
  av_free(picture->data[0]);
  av_free(picture);
  if (tmp_picture) {
    av_free(tmp_picture->data[0]);
    av_free(tmp_picture);
  }
  av_free(video_outbuf);
}

void video::close() {
  av_write_trailer(oc);
  close_video(oc, video_st);
}


void video::copy_image(AVFrame *pict, uint8_t *image) {
  int stride = RGB_STRIDE;
  
  for (int i = 0; i < width*height*stride; i++) {
    pict->data[0][i] = image[i];
  }
}

void video::python_write_frame(boost::python::list &l) {
  int stride = RGB_STRIDE;
  uint8_t *buffer = new uint8_t[width*height*stride];
  
  for (int i = 0; i < width*height*stride; i++)
    buffer[i] = boost::python::extract<uint8_t>(l[i]);
  
  write_frame(buffer);

  delete buffer;
}

void video::write_frame(uint8_t *image) {
  int out_size, ret;
  AVCodecContext *c;
  AVFrame *picture_ptr;
  
  c = &video_st->codec;
  
  if (c->pix_fmt != PIX_FMT_RGB24) {
    /* as we only generate a RGB24 picture, we must convert it
       to the codec pixel format if needed */
    
    copy_image(tmp_picture, image);
    
    img_convert((AVPicture *)picture, c->pix_fmt, 
                (AVPicture *)tmp_picture, PIX_FMT_RGB24,
                c->width, c->height);
    
  } else {
    copy_image(picture, image);
  }
  
  picture_ptr = picture;
  
  if (oc->oformat->flags & AVFMT_RAWPICTURE) {
    /* raw video case. The API will change slightly in the near
       future for that */
    AVPacket pkt;
    av_init_packet(&pkt);
    
    pkt.flags |= PKT_FLAG_KEY;
    pkt.stream_index = video_st->index;
    pkt.data= (uint8_t *)picture_ptr;
    pkt.size= sizeof(AVPicture);
    
    ret = av_write_frame(oc, &pkt);
  } else {
    /* encode the image */
    out_size = avcodec_encode_video(c, video_outbuf, video_outbuf_size, 
                                    picture_ptr);
    /* if zero size, it means the image was buffered */
    if (out_size != 0) {
      AVPacket pkt;
      av_init_packet(&pkt);
      
      pkt.pts= c->coded_frame->pts;
      if(c->coded_frame->key_frame)
        pkt.flags |= PKT_FLAG_KEY;
      pkt.stream_index= video_st->index;
      pkt.data= video_outbuf;
      pkt.size= out_size;
      
      /* write the compressed frame in the media file */
      ret = av_write_frame(oc, &pkt);
    } else {
      ret = 0;
    }
  }
  if (ret != 0) {
    std::cerr << "Error while writing video frame" << std::endl;
    exit(1);
  }
}

using namespace boost::python;
BOOST_PYTHON_MODULE(pyanim)
{
  class_<video>("Video", init<const char *, unsigned, unsigned, int, int>())
    .def("write_frame", &video::python_write_frame)
    .def("close", &video::close)
    ;
}
